/* G13.H - Header file for package for arithmetic in field G(2^4),
reduced modulo x^4 + x + 1.

Version of 19 May 1987.
*/

#ifndef G13
#define G13     int
G13 Add( G13, G13 ) ;
G13 Mult( G13, G13 ) ;
G13 Div( G13, G13 ) ;
#endif
